"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.any_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import tensorflow_metadata.proto.v0.path_pb2
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor = ...

global___LifecycleStage = LifecycleStage
class _LifecycleStage(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[LifecycleStage.V], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor = ...
    UNKNOWN_STAGE = LifecycleStage.V(0)
    PLANNED = LifecycleStage.V(1)
    ALPHA = LifecycleStage.V(2)
    BETA = LifecycleStage.V(3)
    PRODUCTION = LifecycleStage.V(4)
    DEPRECATED = LifecycleStage.V(5)
    DEBUG_ONLY = LifecycleStage.V(6)
class LifecycleStage(metaclass=_LifecycleStage):
    V = typing.NewType('V', builtins.int)
UNKNOWN_STAGE = LifecycleStage.V(0)
PLANNED = LifecycleStage.V(1)
ALPHA = LifecycleStage.V(2)
BETA = LifecycleStage.V(3)
PRODUCTION = LifecycleStage.V(4)
DEPRECATED = LifecycleStage.V(5)
DEBUG_ONLY = LifecycleStage.V(6)

global___FeatureType = FeatureType
class _FeatureType(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[FeatureType.V], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor = ...
    TYPE_UNKNOWN = FeatureType.V(0)
    BYTES = FeatureType.V(1)
    INT = FeatureType.V(2)
    FLOAT = FeatureType.V(3)
    STRUCT = FeatureType.V(4)
class FeatureType(metaclass=_FeatureType):
    V = typing.NewType('V', builtins.int)
TYPE_UNKNOWN = FeatureType.V(0)
BYTES = FeatureType.V(1)
INT = FeatureType.V(2)
FLOAT = FeatureType.V(3)
STRUCT = FeatureType.V(4)

class Schema(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    class TensorRepresentationGroupEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text = ...

        @property
        def value(self) -> global___TensorRepresentationGroup: ...

        def __init__(self,
            *,
            key : typing.Optional[typing.Text] = ...,
            value : typing.Optional[global___TensorRepresentationGroup] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal[u"key",b"key",u"value",b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"key",b"key",u"value",b"value"]) -> None: ...

    FEATURE_FIELD_NUMBER: builtins.int
    SPARSE_FEATURE_FIELD_NUMBER: builtins.int
    WEIGHTED_FEATURE_FIELD_NUMBER: builtins.int
    STRING_DOMAIN_FIELD_NUMBER: builtins.int
    FLOAT_DOMAIN_FIELD_NUMBER: builtins.int
    INT_DOMAIN_FIELD_NUMBER: builtins.int
    DEFAULT_ENVIRONMENT_FIELD_NUMBER: builtins.int
    ANNOTATION_FIELD_NUMBER: builtins.int
    DATASET_CONSTRAINTS_FIELD_NUMBER: builtins.int
    TENSOR_REPRESENTATION_GROUP_FIELD_NUMBER: builtins.int
    default_environment: google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text] = ...

    @property
    def feature(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Feature]: ...

    @property
    def sparse_feature(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SparseFeature]: ...

    @property
    def weighted_feature(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___WeightedFeature]: ...

    @property
    def string_domain(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___StringDomain]: ...

    @property
    def float_domain(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___FloatDomain]: ...

    @property
    def int_domain(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___IntDomain]: ...

    @property
    def annotation(self) -> global___Annotation: ...

    @property
    def dataset_constraints(self) -> global___DatasetConstraints: ...

    @property
    def tensor_representation_group(self) -> google.protobuf.internal.containers.MessageMap[typing.Text, global___TensorRepresentationGroup]: ...

    def __init__(self,
        *,
        feature : typing.Optional[typing.Iterable[global___Feature]] = ...,
        sparse_feature : typing.Optional[typing.Iterable[global___SparseFeature]] = ...,
        weighted_feature : typing.Optional[typing.Iterable[global___WeightedFeature]] = ...,
        string_domain : typing.Optional[typing.Iterable[global___StringDomain]] = ...,
        float_domain : typing.Optional[typing.Iterable[global___FloatDomain]] = ...,
        int_domain : typing.Optional[typing.Iterable[global___IntDomain]] = ...,
        default_environment : typing.Optional[typing.Iterable[typing.Text]] = ...,
        annotation : typing.Optional[global___Annotation] = ...,
        dataset_constraints : typing.Optional[global___DatasetConstraints] = ...,
        tensor_representation_group : typing.Optional[typing.Mapping[typing.Text, global___TensorRepresentationGroup]] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"annotation",b"annotation",u"dataset_constraints",b"dataset_constraints"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"annotation",b"annotation",u"dataset_constraints",b"dataset_constraints",u"default_environment",b"default_environment",u"feature",b"feature",u"float_domain",b"float_domain",u"int_domain",b"int_domain",u"sparse_feature",b"sparse_feature",u"string_domain",b"string_domain",u"tensor_representation_group",b"tensor_representation_group",u"weighted_feature",b"weighted_feature"]) -> None: ...
global___Schema = Schema

class Feature(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    NAME_FIELD_NUMBER: builtins.int
    DEPRECATED_FIELD_NUMBER: builtins.int
    PRESENCE_FIELD_NUMBER: builtins.int
    GROUP_PRESENCE_FIELD_NUMBER: builtins.int
    SHAPE_FIELD_NUMBER: builtins.int
    VALUE_COUNT_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    DOMAIN_FIELD_NUMBER: builtins.int
    INT_DOMAIN_FIELD_NUMBER: builtins.int
    FLOAT_DOMAIN_FIELD_NUMBER: builtins.int
    STRING_DOMAIN_FIELD_NUMBER: builtins.int
    BOOL_DOMAIN_FIELD_NUMBER: builtins.int
    STRUCT_DOMAIN_FIELD_NUMBER: builtins.int
    NATURAL_LANGUAGE_DOMAIN_FIELD_NUMBER: builtins.int
    IMAGE_DOMAIN_FIELD_NUMBER: builtins.int
    MID_DOMAIN_FIELD_NUMBER: builtins.int
    URL_DOMAIN_FIELD_NUMBER: builtins.int
    TIME_DOMAIN_FIELD_NUMBER: builtins.int
    TIME_OF_DAY_DOMAIN_FIELD_NUMBER: builtins.int
    DISTRIBUTION_CONSTRAINTS_FIELD_NUMBER: builtins.int
    ANNOTATION_FIELD_NUMBER: builtins.int
    SKEW_COMPARATOR_FIELD_NUMBER: builtins.int
    DRIFT_COMPARATOR_FIELD_NUMBER: builtins.int
    IN_ENVIRONMENT_FIELD_NUMBER: builtins.int
    NOT_IN_ENVIRONMENT_FIELD_NUMBER: builtins.int
    LIFECYCLE_STAGE_FIELD_NUMBER: builtins.int
    name: typing.Text = ...
    deprecated: builtins.bool = ...
    type: global___FeatureType.V = ...
    domain: typing.Text = ...
    in_environment: google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text] = ...
    not_in_environment: google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text] = ...
    lifecycle_stage: global___LifecycleStage.V = ...

    @property
    def presence(self) -> global___FeaturePresence: ...

    @property
    def group_presence(self) -> global___FeaturePresenceWithinGroup: ...

    @property
    def shape(self) -> global___FixedShape: ...

    @property
    def value_count(self) -> global___ValueCount: ...

    @property
    def int_domain(self) -> global___IntDomain: ...

    @property
    def float_domain(self) -> global___FloatDomain: ...

    @property
    def string_domain(self) -> global___StringDomain: ...

    @property
    def bool_domain(self) -> global___BoolDomain: ...

    @property
    def struct_domain(self) -> global___StructDomain: ...

    @property
    def natural_language_domain(self) -> global___NaturalLanguageDomain: ...

    @property
    def image_domain(self) -> global___ImageDomain: ...

    @property
    def mid_domain(self) -> global___MIDDomain: ...

    @property
    def url_domain(self) -> global___URLDomain: ...

    @property
    def time_domain(self) -> global___TimeDomain: ...

    @property
    def time_of_day_domain(self) -> global___TimeOfDayDomain: ...

    @property
    def distribution_constraints(self) -> global___DistributionConstraints: ...

    @property
    def annotation(self) -> global___Annotation: ...

    @property
    def skew_comparator(self) -> global___FeatureComparator: ...

    @property
    def drift_comparator(self) -> global___FeatureComparator: ...

    def __init__(self,
        *,
        name : typing.Optional[typing.Text] = ...,
        deprecated : typing.Optional[builtins.bool] = ...,
        presence : typing.Optional[global___FeaturePresence] = ...,
        group_presence : typing.Optional[global___FeaturePresenceWithinGroup] = ...,
        shape : typing.Optional[global___FixedShape] = ...,
        value_count : typing.Optional[global___ValueCount] = ...,
        type : typing.Optional[global___FeatureType.V] = ...,
        domain : typing.Optional[typing.Text] = ...,
        int_domain : typing.Optional[global___IntDomain] = ...,
        float_domain : typing.Optional[global___FloatDomain] = ...,
        string_domain : typing.Optional[global___StringDomain] = ...,
        bool_domain : typing.Optional[global___BoolDomain] = ...,
        struct_domain : typing.Optional[global___StructDomain] = ...,
        natural_language_domain : typing.Optional[global___NaturalLanguageDomain] = ...,
        image_domain : typing.Optional[global___ImageDomain] = ...,
        mid_domain : typing.Optional[global___MIDDomain] = ...,
        url_domain : typing.Optional[global___URLDomain] = ...,
        time_domain : typing.Optional[global___TimeDomain] = ...,
        time_of_day_domain : typing.Optional[global___TimeOfDayDomain] = ...,
        distribution_constraints : typing.Optional[global___DistributionConstraints] = ...,
        annotation : typing.Optional[global___Annotation] = ...,
        skew_comparator : typing.Optional[global___FeatureComparator] = ...,
        drift_comparator : typing.Optional[global___FeatureComparator] = ...,
        in_environment : typing.Optional[typing.Iterable[typing.Text]] = ...,
        not_in_environment : typing.Optional[typing.Iterable[typing.Text]] = ...,
        lifecycle_stage : typing.Optional[global___LifecycleStage.V] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"annotation",b"annotation",u"bool_domain",b"bool_domain",u"deprecated",b"deprecated",u"distribution_constraints",b"distribution_constraints",u"domain",b"domain",u"domain_info",b"domain_info",u"drift_comparator",b"drift_comparator",u"float_domain",b"float_domain",u"group_presence",b"group_presence",u"image_domain",b"image_domain",u"int_domain",b"int_domain",u"lifecycle_stage",b"lifecycle_stage",u"mid_domain",b"mid_domain",u"name",b"name",u"natural_language_domain",b"natural_language_domain",u"presence",b"presence",u"presence_constraints",b"presence_constraints",u"shape",b"shape",u"shape_type",b"shape_type",u"skew_comparator",b"skew_comparator",u"string_domain",b"string_domain",u"struct_domain",b"struct_domain",u"time_domain",b"time_domain",u"time_of_day_domain",b"time_of_day_domain",u"type",b"type",u"url_domain",b"url_domain",u"value_count",b"value_count"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"annotation",b"annotation",u"bool_domain",b"bool_domain",u"deprecated",b"deprecated",u"distribution_constraints",b"distribution_constraints",u"domain",b"domain",u"domain_info",b"domain_info",u"drift_comparator",b"drift_comparator",u"float_domain",b"float_domain",u"group_presence",b"group_presence",u"image_domain",b"image_domain",u"in_environment",b"in_environment",u"int_domain",b"int_domain",u"lifecycle_stage",b"lifecycle_stage",u"mid_domain",b"mid_domain",u"name",b"name",u"natural_language_domain",b"natural_language_domain",u"not_in_environment",b"not_in_environment",u"presence",b"presence",u"presence_constraints",b"presence_constraints",u"shape",b"shape",u"shape_type",b"shape_type",u"skew_comparator",b"skew_comparator",u"string_domain",b"string_domain",u"struct_domain",b"struct_domain",u"time_domain",b"time_domain",u"time_of_day_domain",b"time_of_day_domain",u"type",b"type",u"url_domain",b"url_domain",u"value_count",b"value_count"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal[u"domain_info",b"domain_info"]) -> typing_extensions.Literal["domain","int_domain","float_domain","string_domain","bool_domain","struct_domain","natural_language_domain","image_domain","mid_domain","url_domain","time_domain","time_of_day_domain"]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal[u"presence_constraints",b"presence_constraints"]) -> typing_extensions.Literal["presence","group_presence"]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal[u"shape_type",b"shape_type"]) -> typing_extensions.Literal["shape","value_count"]: ...
global___Feature = Feature

class Annotation(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    TAG_FIELD_NUMBER: builtins.int
    COMMENT_FIELD_NUMBER: builtins.int
    EXTRA_METADATA_FIELD_NUMBER: builtins.int
    tag: google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text] = ...
    comment: google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text] = ...

    @property
    def extra_metadata(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[google.protobuf.any_pb2.Any]: ...

    def __init__(self,
        *,
        tag : typing.Optional[typing.Iterable[typing.Text]] = ...,
        comment : typing.Optional[typing.Iterable[typing.Text]] = ...,
        extra_metadata : typing.Optional[typing.Iterable[google.protobuf.any_pb2.Any]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"comment",b"comment",u"extra_metadata",b"extra_metadata",u"tag",b"tag"]) -> None: ...
global___Annotation = Annotation

class NumericValueComparator(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    MIN_FRACTION_THRESHOLD_FIELD_NUMBER: builtins.int
    MAX_FRACTION_THRESHOLD_FIELD_NUMBER: builtins.int
    min_fraction_threshold: builtins.float = ...
    max_fraction_threshold: builtins.float = ...

    def __init__(self,
        *,
        min_fraction_threshold : typing.Optional[builtins.float] = ...,
        max_fraction_threshold : typing.Optional[builtins.float] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"max_fraction_threshold",b"max_fraction_threshold",u"min_fraction_threshold",b"min_fraction_threshold"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"max_fraction_threshold",b"max_fraction_threshold",u"min_fraction_threshold",b"min_fraction_threshold"]) -> None: ...
global___NumericValueComparator = NumericValueComparator

class DatasetConstraints(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    NUM_EXAMPLES_DRIFT_COMPARATOR_FIELD_NUMBER: builtins.int
    NUM_EXAMPLES_VERSION_COMPARATOR_FIELD_NUMBER: builtins.int
    MIN_EXAMPLES_COUNT_FIELD_NUMBER: builtins.int
    min_examples_count: builtins.int = ...

    @property
    def num_examples_drift_comparator(self) -> global___NumericValueComparator: ...

    @property
    def num_examples_version_comparator(self) -> global___NumericValueComparator: ...

    def __init__(self,
        *,
        num_examples_drift_comparator : typing.Optional[global___NumericValueComparator] = ...,
        num_examples_version_comparator : typing.Optional[global___NumericValueComparator] = ...,
        min_examples_count : typing.Optional[builtins.int] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"min_examples_count",b"min_examples_count",u"num_examples_drift_comparator",b"num_examples_drift_comparator",u"num_examples_version_comparator",b"num_examples_version_comparator"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"min_examples_count",b"min_examples_count",u"num_examples_drift_comparator",b"num_examples_drift_comparator",u"num_examples_version_comparator",b"num_examples_version_comparator"]) -> None: ...
global___DatasetConstraints = DatasetConstraints

class FixedShape(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    class Dim(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        SIZE_FIELD_NUMBER: builtins.int
        NAME_FIELD_NUMBER: builtins.int
        size: builtins.int = ...
        name: typing.Text = ...

        def __init__(self,
            *,
            size : typing.Optional[builtins.int] = ...,
            name : typing.Optional[typing.Text] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal[u"name",b"name",u"size",b"size"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"name",b"name",u"size",b"size"]) -> None: ...

    DIM_FIELD_NUMBER: builtins.int

    @property
    def dim(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___FixedShape.Dim]: ...

    def __init__(self,
        *,
        dim : typing.Optional[typing.Iterable[global___FixedShape.Dim]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"dim",b"dim"]) -> None: ...
global___FixedShape = FixedShape

class ValueCount(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    MIN_FIELD_NUMBER: builtins.int
    MAX_FIELD_NUMBER: builtins.int
    min: builtins.int = ...
    max: builtins.int = ...

    def __init__(self,
        *,
        min : typing.Optional[builtins.int] = ...,
        max : typing.Optional[builtins.int] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"max",b"max",u"min",b"min"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"max",b"max",u"min",b"min"]) -> None: ...
global___ValueCount = ValueCount

class WeightedFeature(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    NAME_FIELD_NUMBER: builtins.int
    FEATURE_FIELD_NUMBER: builtins.int
    WEIGHT_FEATURE_FIELD_NUMBER: builtins.int
    LIFECYCLE_STAGE_FIELD_NUMBER: builtins.int
    name: typing.Text = ...
    lifecycle_stage: global___LifecycleStage.V = ...

    @property
    def feature(self) -> tensorflow_metadata.proto.v0.path_pb2.Path: ...

    @property
    def weight_feature(self) -> tensorflow_metadata.proto.v0.path_pb2.Path: ...

    def __init__(self,
        *,
        name : typing.Optional[typing.Text] = ...,
        feature : typing.Optional[tensorflow_metadata.proto.v0.path_pb2.Path] = ...,
        weight_feature : typing.Optional[tensorflow_metadata.proto.v0.path_pb2.Path] = ...,
        lifecycle_stage : typing.Optional[global___LifecycleStage.V] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"feature",b"feature",u"lifecycle_stage",b"lifecycle_stage",u"name",b"name",u"weight_feature",b"weight_feature"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"feature",b"feature",u"lifecycle_stage",b"lifecycle_stage",u"name",b"name",u"weight_feature",b"weight_feature"]) -> None: ...
global___WeightedFeature = WeightedFeature

class SparseFeature(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    class IndexFeature(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        NAME_FIELD_NUMBER: builtins.int
        name: typing.Text = ...

        def __init__(self,
            *,
            name : typing.Optional[typing.Text] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal[u"name",b"name"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"name",b"name"]) -> None: ...

    class ValueFeature(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        NAME_FIELD_NUMBER: builtins.int
        name: typing.Text = ...

        def __init__(self,
            *,
            name : typing.Optional[typing.Text] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal[u"name",b"name"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"name",b"name"]) -> None: ...

    NAME_FIELD_NUMBER: builtins.int
    DEPRECATED_FIELD_NUMBER: builtins.int
    LIFECYCLE_STAGE_FIELD_NUMBER: builtins.int
    PRESENCE_FIELD_NUMBER: builtins.int
    DENSE_SHAPE_FIELD_NUMBER: builtins.int
    INDEX_FEATURE_FIELD_NUMBER: builtins.int
    IS_SORTED_FIELD_NUMBER: builtins.int
    VALUE_FEATURE_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    name: typing.Text = ...
    deprecated: builtins.bool = ...
    lifecycle_stage: global___LifecycleStage.V = ...
    is_sorted: builtins.bool = ...
    type: global___FeatureType.V = ...

    @property
    def presence(self) -> global___FeaturePresence: ...

    @property
    def dense_shape(self) -> global___FixedShape: ...

    @property
    def index_feature(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SparseFeature.IndexFeature]: ...

    @property
    def value_feature(self) -> global___SparseFeature.ValueFeature: ...

    def __init__(self,
        *,
        name : typing.Optional[typing.Text] = ...,
        deprecated : typing.Optional[builtins.bool] = ...,
        lifecycle_stage : typing.Optional[global___LifecycleStage.V] = ...,
        presence : typing.Optional[global___FeaturePresence] = ...,
        dense_shape : typing.Optional[global___FixedShape] = ...,
        index_feature : typing.Optional[typing.Iterable[global___SparseFeature.IndexFeature]] = ...,
        is_sorted : typing.Optional[builtins.bool] = ...,
        value_feature : typing.Optional[global___SparseFeature.ValueFeature] = ...,
        type : typing.Optional[global___FeatureType.V] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"dense_shape",b"dense_shape",u"deprecated",b"deprecated",u"is_sorted",b"is_sorted",u"lifecycle_stage",b"lifecycle_stage",u"name",b"name",u"presence",b"presence",u"type",b"type",u"value_feature",b"value_feature"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"dense_shape",b"dense_shape",u"deprecated",b"deprecated",u"index_feature",b"index_feature",u"is_sorted",b"is_sorted",u"lifecycle_stage",b"lifecycle_stage",u"name",b"name",u"presence",b"presence",u"type",b"type",u"value_feature",b"value_feature"]) -> None: ...
global___SparseFeature = SparseFeature

class DistributionConstraints(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    MIN_DOMAIN_MASS_FIELD_NUMBER: builtins.int
    min_domain_mass: builtins.float = ...

    def __init__(self,
        *,
        min_domain_mass : typing.Optional[builtins.float] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"min_domain_mass",b"min_domain_mass"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"min_domain_mass",b"min_domain_mass"]) -> None: ...
global___DistributionConstraints = DistributionConstraints

class IntDomain(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    NAME_FIELD_NUMBER: builtins.int
    MIN_FIELD_NUMBER: builtins.int
    MAX_FIELD_NUMBER: builtins.int
    IS_CATEGORICAL_FIELD_NUMBER: builtins.int
    name: typing.Text = ...
    min: builtins.int = ...
    max: builtins.int = ...
    is_categorical: builtins.bool = ...

    def __init__(self,
        *,
        name : typing.Optional[typing.Text] = ...,
        min : typing.Optional[builtins.int] = ...,
        max : typing.Optional[builtins.int] = ...,
        is_categorical : typing.Optional[builtins.bool] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"is_categorical",b"is_categorical",u"max",b"max",u"min",b"min",u"name",b"name"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"is_categorical",b"is_categorical",u"max",b"max",u"min",b"min",u"name",b"name"]) -> None: ...
global___IntDomain = IntDomain

class FloatDomain(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    NAME_FIELD_NUMBER: builtins.int
    MIN_FIELD_NUMBER: builtins.int
    MAX_FIELD_NUMBER: builtins.int
    name: typing.Text = ...
    min: builtins.float = ...
    max: builtins.float = ...

    def __init__(self,
        *,
        name : typing.Optional[typing.Text] = ...,
        min : typing.Optional[builtins.float] = ...,
        max : typing.Optional[builtins.float] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"max",b"max",u"min",b"min",u"name",b"name"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"max",b"max",u"min",b"min",u"name",b"name"]) -> None: ...
global___FloatDomain = FloatDomain

class StructDomain(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    FEATURE_FIELD_NUMBER: builtins.int
    SPARSE_FEATURE_FIELD_NUMBER: builtins.int

    @property
    def feature(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Feature]: ...

    @property
    def sparse_feature(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SparseFeature]: ...

    def __init__(self,
        *,
        feature : typing.Optional[typing.Iterable[global___Feature]] = ...,
        sparse_feature : typing.Optional[typing.Iterable[global___SparseFeature]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"feature",b"feature",u"sparse_feature",b"sparse_feature"]) -> None: ...
global___StructDomain = StructDomain

class StringDomain(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    NAME_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    name: typing.Text = ...
    value: google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text] = ...

    def __init__(self,
        *,
        name : typing.Optional[typing.Text] = ...,
        value : typing.Optional[typing.Iterable[typing.Text]] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"name",b"name"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"name",b"name",u"value",b"value"]) -> None: ...
global___StringDomain = StringDomain

class BoolDomain(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    NAME_FIELD_NUMBER: builtins.int
    TRUE_VALUE_FIELD_NUMBER: builtins.int
    FALSE_VALUE_FIELD_NUMBER: builtins.int
    name: typing.Text = ...
    true_value: typing.Text = ...
    false_value: typing.Text = ...

    def __init__(self,
        *,
        name : typing.Optional[typing.Text] = ...,
        true_value : typing.Optional[typing.Text] = ...,
        false_value : typing.Optional[typing.Text] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"false_value",b"false_value",u"name",b"name",u"true_value",b"true_value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"false_value",b"false_value",u"name",b"name",u"true_value",b"true_value"]) -> None: ...
global___BoolDomain = BoolDomain

class NaturalLanguageDomain(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...

    def __init__(self,
        ) -> None: ...
global___NaturalLanguageDomain = NaturalLanguageDomain

class ImageDomain(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...

    def __init__(self,
        ) -> None: ...
global___ImageDomain = ImageDomain

class MIDDomain(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...

    def __init__(self,
        ) -> None: ...
global___MIDDomain = MIDDomain

class URLDomain(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...

    def __init__(self,
        ) -> None: ...
global___URLDomain = URLDomain

class TimeDomain(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    class _IntegerTimeFormat(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[IntegerTimeFormat.V], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor = ...
        FORMAT_UNKNOWN = TimeDomain.IntegerTimeFormat.V(0)
        UNIX_DAYS = TimeDomain.IntegerTimeFormat.V(5)
        UNIX_SECONDS = TimeDomain.IntegerTimeFormat.V(1)
        UNIX_MILLISECONDS = TimeDomain.IntegerTimeFormat.V(2)
        UNIX_MICROSECONDS = TimeDomain.IntegerTimeFormat.V(3)
        UNIX_NANOSECONDS = TimeDomain.IntegerTimeFormat.V(4)
    class IntegerTimeFormat(metaclass=_IntegerTimeFormat):
        V = typing.NewType('V', builtins.int)
    FORMAT_UNKNOWN = TimeDomain.IntegerTimeFormat.V(0)
    UNIX_DAYS = TimeDomain.IntegerTimeFormat.V(5)
    UNIX_SECONDS = TimeDomain.IntegerTimeFormat.V(1)
    UNIX_MILLISECONDS = TimeDomain.IntegerTimeFormat.V(2)
    UNIX_MICROSECONDS = TimeDomain.IntegerTimeFormat.V(3)
    UNIX_NANOSECONDS = TimeDomain.IntegerTimeFormat.V(4)

    STRING_FORMAT_FIELD_NUMBER: builtins.int
    INTEGER_FORMAT_FIELD_NUMBER: builtins.int
    string_format: typing.Text = ...
    integer_format: global___TimeDomain.IntegerTimeFormat.V = ...

    def __init__(self,
        *,
        string_format : typing.Optional[typing.Text] = ...,
        integer_format : typing.Optional[global___TimeDomain.IntegerTimeFormat.V] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"format",b"format",u"integer_format",b"integer_format",u"string_format",b"string_format"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"format",b"format",u"integer_format",b"integer_format",u"string_format",b"string_format"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal[u"format",b"format"]) -> typing_extensions.Literal["string_format","integer_format"]: ...
global___TimeDomain = TimeDomain

class TimeOfDayDomain(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    class _IntegerTimeOfDayFormat(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[IntegerTimeOfDayFormat.V], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor = ...
        FORMAT_UNKNOWN = TimeOfDayDomain.IntegerTimeOfDayFormat.V(0)
        PACKED_64_NANOS = TimeOfDayDomain.IntegerTimeOfDayFormat.V(1)
    class IntegerTimeOfDayFormat(metaclass=_IntegerTimeOfDayFormat):
        V = typing.NewType('V', builtins.int)
    FORMAT_UNKNOWN = TimeOfDayDomain.IntegerTimeOfDayFormat.V(0)
    PACKED_64_NANOS = TimeOfDayDomain.IntegerTimeOfDayFormat.V(1)

    STRING_FORMAT_FIELD_NUMBER: builtins.int
    INTEGER_FORMAT_FIELD_NUMBER: builtins.int
    string_format: typing.Text = ...
    integer_format: global___TimeOfDayDomain.IntegerTimeOfDayFormat.V = ...

    def __init__(self,
        *,
        string_format : typing.Optional[typing.Text] = ...,
        integer_format : typing.Optional[global___TimeOfDayDomain.IntegerTimeOfDayFormat.V] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"format",b"format",u"integer_format",b"integer_format",u"string_format",b"string_format"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"format",b"format",u"integer_format",b"integer_format",u"string_format",b"string_format"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal[u"format",b"format"]) -> typing_extensions.Literal["string_format","integer_format"]: ...
global___TimeOfDayDomain = TimeOfDayDomain

class FeaturePresence(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    MIN_FRACTION_FIELD_NUMBER: builtins.int
    MIN_COUNT_FIELD_NUMBER: builtins.int
    min_fraction: builtins.float = ...
    min_count: builtins.int = ...

    def __init__(self,
        *,
        min_fraction : typing.Optional[builtins.float] = ...,
        min_count : typing.Optional[builtins.int] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"min_count",b"min_count",u"min_fraction",b"min_fraction"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"min_count",b"min_count",u"min_fraction",b"min_fraction"]) -> None: ...
global___FeaturePresence = FeaturePresence

class FeaturePresenceWithinGroup(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    REQUIRED_FIELD_NUMBER: builtins.int
    required: builtins.bool = ...

    def __init__(self,
        *,
        required : typing.Optional[builtins.bool] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"required",b"required"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"required",b"required"]) -> None: ...
global___FeaturePresenceWithinGroup = FeaturePresenceWithinGroup

class InfinityNorm(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    THRESHOLD_FIELD_NUMBER: builtins.int
    threshold: builtins.float = ...

    def __init__(self,
        *,
        threshold : typing.Optional[builtins.float] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"threshold",b"threshold"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"threshold",b"threshold"]) -> None: ...
global___InfinityNorm = InfinityNorm

class FeatureComparator(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    INFINITY_NORM_FIELD_NUMBER: builtins.int

    @property
    def infinity_norm(self) -> global___InfinityNorm: ...

    def __init__(self,
        *,
        infinity_norm : typing.Optional[global___InfinityNorm] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"infinity_norm",b"infinity_norm"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"infinity_norm",b"infinity_norm"]) -> None: ...
global___FeatureComparator = FeatureComparator

class TensorRepresentation(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    class DefaultValue(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        FLOAT_VALUE_FIELD_NUMBER: builtins.int
        INT_VALUE_FIELD_NUMBER: builtins.int
        BYTES_VALUE_FIELD_NUMBER: builtins.int
        UINT_VALUE_FIELD_NUMBER: builtins.int
        float_value: builtins.float = ...
        int_value: builtins.int = ...
        bytes_value: builtins.bytes = ...
        uint_value: builtins.int = ...

        def __init__(self,
            *,
            float_value : typing.Optional[builtins.float] = ...,
            int_value : typing.Optional[builtins.int] = ...,
            bytes_value : typing.Optional[builtins.bytes] = ...,
            uint_value : typing.Optional[builtins.int] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal[u"bytes_value",b"bytes_value",u"float_value",b"float_value",u"int_value",b"int_value",u"kind",b"kind",u"uint_value",b"uint_value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"bytes_value",b"bytes_value",u"float_value",b"float_value",u"int_value",b"int_value",u"kind",b"kind",u"uint_value",b"uint_value"]) -> None: ...
        def WhichOneof(self, oneof_group: typing_extensions.Literal[u"kind",b"kind"]) -> typing_extensions.Literal["float_value","int_value","bytes_value","uint_value"]: ...

    class DenseTensor(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        COLUMN_NAME_FIELD_NUMBER: builtins.int
        SHAPE_FIELD_NUMBER: builtins.int
        DEFAULT_VALUE_FIELD_NUMBER: builtins.int
        column_name: typing.Text = ...

        @property
        def shape(self) -> global___FixedShape: ...

        @property
        def default_value(self) -> global___TensorRepresentation.DefaultValue: ...

        def __init__(self,
            *,
            column_name : typing.Optional[typing.Text] = ...,
            shape : typing.Optional[global___FixedShape] = ...,
            default_value : typing.Optional[global___TensorRepresentation.DefaultValue] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal[u"column_name",b"column_name",u"default_value",b"default_value",u"shape",b"shape"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"column_name",b"column_name",u"default_value",b"default_value",u"shape",b"shape"]) -> None: ...

    class VarLenSparseTensor(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        COLUMN_NAME_FIELD_NUMBER: builtins.int
        column_name: typing.Text = ...

        def __init__(self,
            *,
            column_name : typing.Optional[typing.Text] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal[u"column_name",b"column_name"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"column_name",b"column_name"]) -> None: ...

    class SparseTensor(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        DENSE_SHAPE_FIELD_NUMBER: builtins.int
        INDEX_COLUMN_NAMES_FIELD_NUMBER: builtins.int
        VALUE_COLUMN_NAME_FIELD_NUMBER: builtins.int
        index_column_names: google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text] = ...
        value_column_name: typing.Text = ...

        @property
        def dense_shape(self) -> global___FixedShape: ...

        def __init__(self,
            *,
            dense_shape : typing.Optional[global___FixedShape] = ...,
            index_column_names : typing.Optional[typing.Iterable[typing.Text]] = ...,
            value_column_name : typing.Optional[typing.Text] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal[u"dense_shape",b"dense_shape",u"value_column_name",b"value_column_name"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"dense_shape",b"dense_shape",u"index_column_names",b"index_column_names",u"value_column_name",b"value_column_name"]) -> None: ...

    DENSE_TENSOR_FIELD_NUMBER: builtins.int
    VARLEN_SPARSE_TENSOR_FIELD_NUMBER: builtins.int
    SPARSE_TENSOR_FIELD_NUMBER: builtins.int

    @property
    def dense_tensor(self) -> global___TensorRepresentation.DenseTensor: ...

    @property
    def varlen_sparse_tensor(self) -> global___TensorRepresentation.VarLenSparseTensor: ...

    @property
    def sparse_tensor(self) -> global___TensorRepresentation.SparseTensor: ...

    def __init__(self,
        *,
        dense_tensor : typing.Optional[global___TensorRepresentation.DenseTensor] = ...,
        varlen_sparse_tensor : typing.Optional[global___TensorRepresentation.VarLenSparseTensor] = ...,
        sparse_tensor : typing.Optional[global___TensorRepresentation.SparseTensor] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"dense_tensor",b"dense_tensor",u"kind",b"kind",u"sparse_tensor",b"sparse_tensor",u"varlen_sparse_tensor",b"varlen_sparse_tensor"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"dense_tensor",b"dense_tensor",u"kind",b"kind",u"sparse_tensor",b"sparse_tensor",u"varlen_sparse_tensor",b"varlen_sparse_tensor"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal[u"kind",b"kind"]) -> typing_extensions.Literal["dense_tensor","varlen_sparse_tensor","sparse_tensor"]: ...
global___TensorRepresentation = TensorRepresentation

class TensorRepresentationGroup(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    class TensorRepresentationEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text = ...

        @property
        def value(self) -> global___TensorRepresentation: ...

        def __init__(self,
            *,
            key : typing.Optional[typing.Text] = ...,
            value : typing.Optional[global___TensorRepresentation] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal[u"key",b"key",u"value",b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"key",b"key",u"value",b"value"]) -> None: ...

    TENSOR_REPRESENTATION_FIELD_NUMBER: builtins.int

    @property
    def tensor_representation(self) -> google.protobuf.internal.containers.MessageMap[typing.Text, global___TensorRepresentation]: ...

    def __init__(self,
        *,
        tensor_representation : typing.Optional[typing.Mapping[typing.Text, global___TensorRepresentation]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"tensor_representation",b"tensor_representation"]) -> None: ...
global___TensorRepresentationGroup = TensorRepresentationGroup



import random
import string
import tempfile
from pathlib import Path
from textwrap import dedent

import pytest

from feast.feature_store import FeatureStore
from tests.cli_utils import CliRunner
from tests.online_read_write_test import basic_rw_test


@pytest.mark.integration
def test_basic() -> None:
    project_id = "".join(
        random.choice(string.ascii_lowercase + string.digits) for _ in range(10)
    )
    runner = CliRunner()
    with tempfile.TemporaryDirectory() as repo_dir_name, tempfile.TemporaryDirectory() as data_dir_name:

        repo_path = Path(repo_dir_name)
        data_path = Path(data_dir_name)

        repo_config = repo_path / "feature_store.yaml"

        repo_config.write_text(
            dedent(
                f"""
        project: {project_id}
        registry: {data_path / "registry.db"}
        provider: gcp
        """
            )
        )

        repo_example = repo_path / "example.py"
        repo_example.write_text(
            (Path(__file__).parent / "example_feature_repo_1.py").read_text()
        )

        result = runner.run(["apply"], cwd=repo_path)
        assert result.returncode == 0

        # Doing another apply should be a no op, and should not cause errors
        result = runner.run(["apply"], cwd=repo_path)
        assert result.returncode == 0

        basic_rw_test(
            FeatureStore(repo_path=str(repo_path), config=None),
            view_name="driver_locations",
        )

        result = runner.run(["teardown"], cwd=repo_path)
        assert result.returncode == 0



import os
import random
import string
import time
from datetime import datetime, timedelta
from tempfile import TemporaryDirectory

import numpy as np
import pandas as pd
import pytest
from google.cloud import bigquery
from pandas.testing import assert_frame_equal
from pytz import utc

import feast.driver_test_data as driver_data
from feast import utils
from feast.data_source import BigQuerySource, FileSource
from feast.entity import Entity
from feast.feature import Feature
from feast.feature_store import FeatureStore
from feast.feature_view import FeatureView
from feast.infra.provider import ENTITY_DF_EVENT_TIMESTAMP_COL
from feast.repo_config import RepoConfig, SqliteOnlineStoreConfig
from feast.value_type import ValueType

np.random.seed(0)

PROJECT_NAME = "default"


def generate_entities(date):
    end_date = date
    before_start_date = end_date - timedelta(days=14)
    start_date = end_date - timedelta(days=7)
    after_end_date = end_date + timedelta(days=7)
    customer_entities = [1001, 1002, 1003, 1004, 1005, 1006, 1007, 1008, 1009, 1010]
    driver_entities = [5001, 5002, 5003, 5004, 5005, 5006, 5007, 5008, 5009, 5010]
    orders_df = driver_data.create_orders_df(
        customer_entities, driver_entities, before_start_date, after_end_date, 20
    )
    return customer_entities, driver_entities, end_date, orders_df, start_date


def stage_driver_hourly_stats_parquet_source(directory, df):
    # Write to disk
    driver_stats_path = os.path.join(directory, "driver_stats.parquet")
    df.to_parquet(path=driver_stats_path, allow_truncated_timestamps=True)
    return FileSource(
        path=driver_stats_path,
        event_timestamp_column="datetime",
        created_timestamp_column="",
    )


def stage_driver_hourly_stats_bigquery_source(df, table_id):
    client = bigquery.Client()
    job_config = bigquery.LoadJobConfig()
    df.reset_index(drop=True, inplace=True)
    job = client.load_table_from_dataframe(df, table_id, job_config=job_config)
    job.result()


def create_driver_hourly_stats_feature_view(source):
    driver_stats_feature_view = FeatureView(
        name="driver_stats",
        entities=["driver"],
        features=[
            Feature(name="conv_rate", dtype=ValueType.FLOAT),
            Feature(name="acc_rate", dtype=ValueType.FLOAT),
            Feature(name="avg_daily_trips", dtype=ValueType.INT32),
        ],
        input=source,
        ttl=timedelta(hours=2),
    )
    return driver_stats_feature_view


def stage_customer_daily_profile_parquet_source(directory, df):
    customer_profile_path = os.path.join(directory, "customer_profile.parquet")
    df.to_parquet(path=customer_profile_path, allow_truncated_timestamps=True)
    return FileSource(
        path=customer_profile_path,
        event_timestamp_column="datetime",
        created_timestamp_column="created",
    )


def stage_customer_daily_profile_bigquery_source(df, table_id):
    client = bigquery.Client()
    job_config = bigquery.LoadJobConfig()
    df.reset_index(drop=True, inplace=True)
    job = client.load_table_from_dataframe(df, table_id, job_config=job_config)
    job.result()


def create_customer_daily_profile_feature_view(source):
    customer_profile_feature_view = FeatureView(
        name="customer_profile",
        entities=["customer_id"],
        features=[
            Feature(name="current_balance", dtype=ValueType.FLOAT),
            Feature(name="avg_passenger_count", dtype=ValueType.FLOAT),
            Feature(name="lifetime_trip_count", dtype=ValueType.INT32),
        ],
        input=source,
        ttl=timedelta(days=2),
    )
    return customer_profile_feature_view


# Converts the given column of the pandas records to UTC timestamps
def convert_timestamp_records_to_utc(records, column):
    for record in records:
        record[column] = utils.make_tzaware(record[column]).astimezone(utc)
    return records


# Find the latest record in the given time range and filter
def find_asof_record(records, ts_key, ts_start, ts_end, filter_key, filter_value):
    found_record = {}
    for record in records:
        if record[filter_key] == filter_value and ts_start <= record[ts_key] <= ts_end:
            if not found_record or found_record[ts_key] < record[ts_key]:
                found_record = record
    return found_record


def get_expected_training_df(
    customer_df: pd.DataFrame,
    customer_fv: FeatureView,
    driver_df: pd.DataFrame,
    driver_fv: FeatureView,
    orders_df: pd.DataFrame,
):
    # Convert all pandas dataframes into records with UTC timestamps
    order_records = convert_timestamp_records_to_utc(
        orders_df.to_dict("records"), "event_timestamp"
    )
    driver_records = convert_timestamp_records_to_utc(
        driver_df.to_dict("records"), driver_fv.input.event_timestamp_column
    )
    customer_records = convert_timestamp_records_to_utc(
        customer_df.to_dict("records"), customer_fv.input.event_timestamp_column
    )

    # Manually do point-in-time join of orders to drivers and customers records
    for order_record in order_records:
        driver_record = find_asof_record(
            driver_records,
            ts_key=driver_fv.input.event_timestamp_column,
            ts_start=order_record["event_timestamp"] - driver_fv.ttl,
            ts_end=order_record["event_timestamp"],
            filter_key="driver_id",
            filter_value=order_record["driver_id"],
        )
        customer_record = find_asof_record(
            customer_records,
            ts_key=customer_fv.input.event_timestamp_column,
            ts_start=order_record["event_timestamp"] - customer_fv.ttl,
            ts_end=order_record["event_timestamp"],
            filter_key="customer_id",
            filter_value=order_record["customer_id"],
        )
        order_record.update(
            {
                f"driver_stats__{k}": driver_record.get(k, None)
                for k in ("conv_rate", "avg_daily_trips")
            }
        )
        order_record.update(
            {
                f"customer_profile__{k}": customer_record.get(k, None)
                for k in (
                    "current_balance",
                    "avg_passenger_count",
                    "lifetime_trip_count",
                )
            }
        )

    # Convert records back to pandas dataframe
    expected_df = pd.DataFrame(order_records)

    # Move "datetime" column to front
    current_cols = expected_df.columns.tolist()
    current_cols.remove(ENTITY_DF_EVENT_TIMESTAMP_COL)
    expected_df = expected_df[[ENTITY_DF_EVENT_TIMESTAMP_COL] + current_cols]

    # Cast some columns to expected types, since we lose information when converting pandas DFs into Python objects.
    expected_df["order_is_success"] = expected_df["order_is_success"].astype("int32")
    expected_df["customer_profile__current_balance"] = expected_df[
        "customer_profile__current_balance"
    ].astype("float32")
    expected_df["customer_profile__avg_passenger_count"] = expected_df[
        "customer_profile__avg_passenger_count"
    ].astype("float32")

    return expected_df


def stage_orders_bigquery(df, table_id):
    client = bigquery.Client()
    job_config = bigquery.LoadJobConfig()
    df.reset_index(drop=True, inplace=True)
    job = client.load_table_from_dataframe(df, table_id, job_config=job_config)
    job.result()


class BigQueryDataSet:
    def __init__(self, dataset_name):
        self.name = dataset_name

    def __enter__(self):
        client = bigquery.Client()
        dataset = bigquery.Dataset(f"{client.project}.{self.name}")
        dataset.location = "US"
        dataset = client.create_dataset(dataset, exists_ok=True)
        return dataset

    def __exit__(self, exc_type, exc_value, exc_traceback):
        print("Tearing down BigQuery dataset")
        client = bigquery.Client()
        dataset_id = f"{client.project}.{self.name}"

        client.delete_dataset(dataset_id, delete_contents=True, not_found_ok=True)
        print(f"Deleted dataset '{dataset_id}'")
        if exc_type:
            print(
                "***Logging exception {}***".format(
                    (exc_type, exc_value, exc_traceback)
                )
            )


def test_historical_features_from_parquet_sources():
    start_date = datetime.now().replace(microsecond=0, second=0, minute=0)
    (
        customer_entities,
        driver_entities,
        end_date,
        orders_df,
        start_date,
    ) = generate_entities(start_date)

    with TemporaryDirectory() as temp_dir:
        driver_df = driver_data.create_driver_hourly_stats_df(
            driver_entities, start_date, end_date
        )
        driver_source = stage_driver_hourly_stats_parquet_source(temp_dir, driver_df)
        driver_fv = create_driver_hourly_stats_feature_view(driver_source)
        customer_df = driver_data.create_customer_daily_profile_df(
            customer_entities, start_date, end_date
        )
        customer_source = stage_customer_daily_profile_parquet_source(
            temp_dir, customer_df
        )
        customer_fv = create_customer_daily_profile_feature_view(customer_source)
        driver = Entity(name="driver", join_key="driver_id", value_type=ValueType.INT64)
        customer = Entity(name="customer_id", value_type=ValueType.INT64)

        store = FeatureStore(
            config=RepoConfig(
                registry=os.path.join(temp_dir, "registry.db"),
                project="default",
                provider="local",
                online_store=SqliteOnlineStoreConfig(
                    path=os.path.join(temp_dir, "online_store.db")
                ),
            )
        )

        store.apply([driver, customer, driver_fv, customer_fv])

        job = store.get_historical_features(
            entity_df=orders_df,
            feature_refs=[
                "driver_stats:conv_rate",
                "driver_stats:avg_daily_trips",
                "customer_profile:current_balance",
                "customer_profile:avg_passenger_count",
                "customer_profile:lifetime_trip_count",
            ],
        )

        actual_df = job.to_df()
        expected_df = get_expected_training_df(
            customer_df, customer_fv, driver_df, driver_fv, orders_df,
        )
        assert_frame_equal(
            expected_df.sort_values(
                by=[
                    ENTITY_DF_EVENT_TIMESTAMP_COL,
                    "order_id",
                    "driver_id",
                    "customer_id",
                ]
            ).reset_index(drop=True),
            actual_df.sort_values(
                by=[
                    ENTITY_DF_EVENT_TIMESTAMP_COL,
                    "order_id",
                    "driver_id",
                    "customer_id",
                ]
            ).reset_index(drop=True),
        )


@pytest.mark.integration
@pytest.mark.parametrize(
    "provider_type", ["local", "gcp"],
)
def test_historical_features_from_bigquery_sources(provider_type):
    start_date = datetime.now().replace(microsecond=0, second=0, minute=0)
    (
        customer_entities,
        driver_entities,
        end_date,
        orders_df,
        start_date,
    ) = generate_entities(start_date)

    # bigquery_dataset = "test_hist_retrieval_static"
    bigquery_dataset = f"test_hist_retrieval_{int(time.time())}"

    with BigQueryDataSet(bigquery_dataset), TemporaryDirectory() as temp_dir:
        gcp_project = bigquery.Client().project

        # Orders Query
        table_id = f"{bigquery_dataset}.orders"
        stage_orders_bigquery(orders_df, table_id)
        entity_df_query = f"SELECT * FROM {gcp_project}.{table_id}"

        # Driver Feature View
        driver_df = driver_data.create_driver_hourly_stats_df(
            driver_entities, start_date, end_date
        )
        driver_table_id = f"{gcp_project}.{bigquery_dataset}.driver_hourly"
        stage_driver_hourly_stats_bigquery_source(driver_df, driver_table_id)
        driver_source = BigQuerySource(
            table_ref=driver_table_id,
            event_timestamp_column="datetime",
            created_timestamp_column="created",
        )
        driver_fv = create_driver_hourly_stats_feature_view(driver_source)

        # Customer Feature View
        customer_df = driver_data.create_customer_daily_profile_df(
            customer_entities, start_date, end_date
        )
        customer_table_id = f"{gcp_project}.{bigquery_dataset}.customer_profile"

        stage_customer_daily_profile_bigquery_source(customer_df, customer_table_id)
        customer_source = BigQuerySource(
            table_ref=customer_table_id,
            event_timestamp_column="datetime",
            created_timestamp_column="",
        )
        customer_fv = create_customer_daily_profile_feature_view(customer_source)

        driver = Entity(name="driver", join_key="driver_id", value_type=ValueType.INT64)
        customer = Entity(name="customer_id", value_type=ValueType.INT64)

        if provider_type == "local":
            store = FeatureStore(
                config=RepoConfig(
                    registry=os.path.join(temp_dir, "registry.db"),
                    project="default",
                    provider="local",
                    online_store=SqliteOnlineStoreConfig(
                        path=os.path.join(temp_dir, "online_store.db"),
                    ),
                )
            )
        elif provider_type == "gcp":
            store = FeatureStore(
                config=RepoConfig(
                    registry=os.path.join(temp_dir, "registry.db"),
                    project="".join(
                        random.choices(string.ascii_uppercase + string.digits, k=10)
                    ),
                    provider="gcp",
                )
            )
        else:
            raise Exception("Invalid provider used as part of test configuration")

        store.apply([driver, customer, driver_fv, customer_fv])

        expected_df = get_expected_training_df(
            customer_df, customer_fv, driver_df, driver_fv, orders_df,
        )

        job_from_sql = store.get_historical_features(
            entity_df=entity_df_query,
            feature_refs=[
                "driver_stats:conv_rate",
                "driver_stats:avg_daily_trips",
                "customer_profile:current_balance",
                "customer_profile:avg_passenger_count",
                "customer_profile:lifetime_trip_count",
            ],
        )
        actual_df_from_sql_entities = job_from_sql.to_df()

        assert_frame_equal(
            expected_df.sort_values(
                by=[
                    ENTITY_DF_EVENT_TIMESTAMP_COL,
                    "order_id",
                    "driver_id",
                    "customer_id",
                ]
            ).reset_index(drop=True),
            actual_df_from_sql_entities.sort_values(
                by=[
                    ENTITY_DF_EVENT_TIMESTAMP_COL,
                    "order_id",
                    "driver_id",
                    "customer_id",
                ]
            ).reset_index(drop=True),
            check_dtype=False,
        )

        job_from_df = store.get_historical_features(
            entity_df=orders_df,
            feature_refs=[
                "driver_stats:conv_rate",
                "driver_stats:avg_daily_trips",
                "customer_profile:current_balance",
                "customer_profile:avg_passenger_count",
                "customer_profile:lifetime_trip_count",
            ],
        )
        actual_df_from_df_entities = job_from_df.to_df()

        assert_frame_equal(
            expected_df.sort_values(
                by=[
                    ENTITY_DF_EVENT_TIMESTAMP_COL,
                    "order_id",
                    "driver_id",
                    "customer_id",
                ]
            ).reset_index(drop=True),
            actual_df_from_df_entities.sort_values(
                by=[
                    ENTITY_DF_EVENT_TIMESTAMP_COL,
                    "order_id",
                    "driver_id",
                    "customer_id",
                ]
            ).reset_index(drop=True),
            check_dtype=False,
        )



# Copyright 2019 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import multiprocessing
from sys import platform

import pytest


def pytest_configure(config):
    if platform in ["darwin", "windows"]:
        multiprocessing.set_start_method("spawn")
    else:
        multiprocessing.set_start_method("fork")
    config.addinivalue_line(
        "markers", "integration: mark test that has external dependencies"
    )


def pytest_addoption(parser):
    parser.addoption(
        "--integration",
        action="store_true",
        default=False,
        help="Run tests with external dependencies",
    )


def pytest_collection_modifyitems(config, items):
    if config.getoption("--integration"):
        return
    skip_integration = pytest.mark.skip(
        reason="not running tests with external dependencies"
    )
    for item in items:
        if "integration" in item.keywords:
            item.add_marker(skip_integration)



import random
import string
import subprocess
import sys
import tempfile
from contextlib import contextmanager
from pathlib import Path
from textwrap import dedent
from typing import List

from feast import cli
from feast.feature_store import FeatureStore


def get_example_repo(example_repo_py) -> str:
    return (Path(__file__).parent / example_repo_py).read_text()


class CliRunner:
    """
    NB. We can't use test runner helper from click here, since it doesn't start a new Python
    interpreter. And we need a new interpreter for each test since we dynamically import
    modules from the feature repo, and it is hard to clean up that state otherwise.
    """

    def run(self, args: List[str], cwd: Path) -> subprocess.CompletedProcess:
        return subprocess.run([sys.executable, cli.__file__] + args, cwd=cwd)

    @contextmanager
    def local_repo(self, example_repo_py: str):
        """
        Convenience method to set up all the boilerplate for a local feature repo.
        """
        project_id = "test" + "".join(
            random.choice(string.ascii_lowercase + string.digits) for _ in range(10)
        )

        with tempfile.TemporaryDirectory() as repo_dir_name, tempfile.TemporaryDirectory() as data_dir_name:

            repo_path = Path(repo_dir_name)
            data_path = Path(data_dir_name)

            repo_config = repo_path / "feature_store.yaml"

            repo_config.write_text(
                dedent(
                    f"""
            project: {project_id}
            registry: {data_path / "registry.db"}
            provider: local
            online_store:
                path: {data_path / "online_store.db"}
            """
                )
            )

            repo_example = repo_path / "example.py"
            repo_example.write_text(example_repo_py)

            result = self.run(["apply"], cwd=repo_path)
            assert result.returncode == 0

            yield FeatureStore(repo_path=str(repo_path), config=None)

            result = self.run(["teardown"], cwd=repo_path)
            assert result.returncode == 0



import tempfile
from datetime import datetime, timedelta
from pathlib import Path

from tests.cli_utils import CliRunner


def test_repo_init() -> None:
    """
    This test simply makes sure that you can run `feast apply && feast materialize` on
    the repo created by "feast init" without errors.
    """
    runner = CliRunner()
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)
        result = runner.run(["init", "my_project"], cwd=temp_path)
        repo_path = temp_path / "my_project"
        assert result.returncode == 0
        result = runner.run(["apply"], cwd=repo_path)
        assert result.returncode == 0

        end_date = datetime.utcnow()
        start_date = end_date - timedelta(days=100)
        result = runner.run(
            ["materialize", start_date.isoformat(), end_date.isoformat()], cwd=repo_path
        )
        assert result.returncode == 0



import os
import tempfile
import uuid
from datetime import datetime, timedelta
from typing import List
from urllib.parse import urlparse

import pandas as pd

from feast.config import Config
from feast.data_format import ParquetFormat
from feast.data_source import BigQuerySource, FileSource
from feast.staging.storage_client import get_staging_client

try:
    from google.cloud import bigquery
except ImportError:
    bigquery = None


def stage_entities_to_fs(
    entity_source: pd.DataFrame, staging_location: str, config: Config
) -> FileSource:
    """
    Dumps given (entities) dataframe as parquet file and stage it to remote file storage (subdirectory of staging_location)

    :return: FileSource with remote destination path
    """
    entity_staging_uri = urlparse(os.path.join(staging_location, str(uuid.uuid4())))
    staging_client = get_staging_client(entity_staging_uri.scheme, config)
    with tempfile.NamedTemporaryFile() as df_export_path:
        # prevent casting ns -> ms exception inside pyarrow
        entity_source["event_timestamp"] = entity_source["event_timestamp"].dt.floor(
            "ms"
        )

        entity_source.to_parquet(df_export_path.name)

        with open(df_export_path.name, "rb") as f:
            staging_client.upload_fileobj(
                f, df_export_path.name, remote_uri=entity_staging_uri
            )

    # ToDo: support custom event_timestamp_column
    return FileSource(
        event_timestamp_column="event_timestamp",
        file_format=ParquetFormat(),
        file_url=entity_staging_uri.geturl(),
    )


def table_reference_from_string(table_ref: str):
    """
    Parses reference string with format "{project}:{dataset}.{table}" into bigquery.TableReference
    """
    project, dataset_and_table = table_ref.split(":")
    dataset, table_id = dataset_and_table.split(".")
    return bigquery.TableReference(
        bigquery.DatasetReference(project, dataset), table_id
    )


def stage_entities_to_bq(
    entity_source: pd.DataFrame, project: str, dataset: str
) -> BigQuerySource:
    """
    Stores given (entity) dataframe as new table in BQ. Name of the table generated based on current time.
    Table will expire in 1 day.
    Returns BigQuerySource with reference to created table.
    """
    bq_client = bigquery.Client()
    destination = bigquery.TableReference(
        bigquery.DatasetReference(project, dataset),
        f"_entities_{datetime.now():%Y%m%d%H%M%s}",
    )

    # prevent casting ns -> ms exception inside pyarrow
    entity_source["event_timestamp"] = entity_source["event_timestamp"].dt.floor("ms")

    load_job: bigquery.LoadJob = bq_client.load_table_from_dataframe(
        entity_source, destination
    )
    load_job.result()  # wait until complete

    dest_table: bigquery.Table = bq_client.get_table(destination)
    dest_table.expires = datetime.now() + timedelta(days=1)
    bq_client.update_table(dest_table, fields=["expires"])

    return BigQuerySource(
        event_timestamp_column="event_timestamp",
        table_ref=f"{destination.project}:{destination.dataset_id}.{destination.table_id}",
    )


JOIN_TEMPLATE = """SELECT
  source.*
FROM
  `{entities.project}.{entities.dataset_id}.{entities.table_id}` entities
JOIN
  `{source.project}.{source.dataset_id}.{source.table_id}` source
ON
  ({entity_key})"""


def create_bq_view_of_joined_features_and_entities(
    source: BigQuerySource, entity_source: BigQuerySource, entity_names: List[str]
) -> BigQuerySource:
    """
    Creates BQ view that joins tables from `source` and `entity_source` with join key derived from `entity_names`.
    Returns BigQuerySource with reference to created view.
    """
    bq_client = bigquery.Client()

    source_ref = table_reference_from_string(source.bigquery_options.table_ref)
    entities_ref = table_reference_from_string(entity_source.bigquery_options.table_ref)

    destination_ref = bigquery.TableReference(
        bigquery.DatasetReference(source_ref.project, source_ref.dataset_id),
        f"_view_{source_ref.table_id}_{datetime.now():%Y%m%d%H%M%s}",
    )

    view = bigquery.Table(destination_ref)
    view.view_query = JOIN_TEMPLATE.format(
        entities=entities_ref,
        source=source_ref,
        entity_key=" AND ".join([f"source.{e} = entities.{e}" for e in entity_names]),
    )
    view.expires = datetime.now() + timedelta(days=1)
    bq_client.create_table(view)

    return BigQuerySource(
        event_timestamp_column=source.event_timestamp_column,
        created_timestamp_column=source.created_timestamp_column,
        table_ref=f"{view.project}:{view.dataset_id}.{view.table_id}",
        field_mapping=source.field_mapping,
        date_partition_column=source.date_partition_column,
    )



# Copyright 2020 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from tempfile import mkstemp

import pytest

from feast.config import Config


class TestConfig:
    @pytest.fixture
    def normal_config(self):
        fd, path = mkstemp()
        return Config(path=path)

    def test_init_config_file_with_path(self):
        configuration_string = "[general]\nCORE_URL = grpc://127.0.0.1:6565"

        fd, path = mkstemp()
        with open(fd, "w") as f:
            f.write(configuration_string)
        config = Config(path=path)
        assert config.get("core_url") == "grpc://127.0.0.1:6565"

    def test_load_environmental_variable(self, normal_config):
        import os

        serving_url = "http://196.25.1.1"
        os.environ["FEAST_SERVING_URL"] = serving_url
        assert normal_config.get("SERVING_URL") == serving_url
        del os.environ["FEAST_SERVING_URL"]

    def test_env_var_not_case_sensitive(self, normal_config):
        import os

        serving_url = "http://196.25.1.1"
        os.environ["FEAST_SerVING_url"] = serving_url
        assert normal_config.get("SERVING_URL") == serving_url

    def test_force_options(self):
        fd, path = mkstemp()
        options = {"feast_config_1": "one", "random_config_two": 2}
        config = Config(options, path)
        assert config.get("feast_config_1") == "one"

    def test_init_options_precedence(self):
        """
        Init options > env var > file options > default options
        """
        fd, path = mkstemp()
        os.environ["FEAST_CORE_URL"] = "env"
        options = {"core_url": "init", "serving_url": "init"}
        configuration_string = "[general]\nCORE_URL = file\n"
        with open(fd, "w") as f:
            f.write(configuration_string)
        config = Config(options, path)
        assert config.get("core_url") == "init"
        del os.environ["FEAST_CORE_URL"]

    def test_env_var_precedence(self):
        """
        Env vars > file options > default options
        """
        fd, path = mkstemp()
        os.environ["FEAST_CORE_URL"] = "env"
        configuration_string = "[general]\nCORE_URL = file\n"
        with open(fd, "w") as f:
            f.write(configuration_string)
        config = Config(path=path)
        assert config.get("CORE_URL") == "env"

        del os.environ["FEAST_CORE_URL"]

    def test_file_option_precedence(self):
        """
        file options > default options
        """
        fd, path = mkstemp()
        configuration_string = "[general]\nCORE_URL = file\n"
        with open(fd, "w") as f:
            f.write(configuration_string)
        config = Config(path=path)
        assert config.get("CORE_URL") == "file"

    def test_default_options(self):
        """
        default options
        """
        fd, path = mkstemp()
        config = Config(path=path)
        assert config.get("CORE_URL") == "localhost:6565"

    def test_defaults_are_not_written(self):
        """
        default values are not written to config file
        """
        fd, path = mkstemp()
        config = Config(path=path)
        config.set("option", "value")
        config.save()
        with open(path) as f:
            assert f.read() == "[general]\noption = value\n\n"

    def test_type_casting(self):
        """
        Test type casting of strings to other types
        """
        fd, path = mkstemp()
        os.environ["FEAST_INT_VAR"] = "1"
        os.environ["FEAST_FLOAT_VAR"] = "1.0"
        os.environ["FEAST_BOOLEAN_VAR"] = "True"
        config = Config(path=path)

        assert config.getint("INT_VAR") == 1
        assert config.getfloat("FLOAT_VAR") == 1.0
        assert config.getboolean("BOOLEAN_VAR") is True

    def test_type_casting_of_defaults(self):
        """
        default values are casted as expected
        """
        fd, path = mkstemp()
        config = Config(path=path)
        assert isinstance(config.getboolean("enable_auth"), bool)
        assert isinstance(config.getint("DATAPROC_EXECUTOR_INSTANCES"), int)
        assert isinstance(config.getfloat("DATAPROC_EXECUTOR_INSTANCES"), float)

    def test_set_value(self):
        """
        Test type casting of strings to other types
        """
        fd, path = mkstemp()
        config = Config(path=path)
        config.set("my_val", 1)

        assert config.getint("my_val") == 1

    def test_exists(self):
        """
        Test type casting of strings to other types
        """
        fd, path = mkstemp()
        config = Config(path=path)
        config.set("my_val_exist", 1)

        assert config.exists("my_val_exist") is True
        assert config.exists("my_val_not_exist") is False



#
#  Copyright 2019 The Feast Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
import logging
import os
from configparser import ConfigParser, NoOptionError
from os.path import expanduser, join
from typing import Dict, Optional

from feast.constants import (
    CONFIG_FEAST_ENV_VAR_PREFIX,
    CONFIG_FILE_DEFAULT_DIRECTORY,
    CONFIG_FILE_NAME,
    CONFIG_FILE_SECTION,
    FEAST_CONFIG_FILE_ENV,
)
from feast.constants import ConfigOptions as opt

_logger = logging.getLogger(__name__)
_UNSET = object()


def _init_config(path: str):
    """
    Returns a ConfigParser that reads in a feast configuration file. If the
    file does not exist it will be created.

    Args:
        path: Optional path to initialize as Feast configuration

    Returns: ConfigParser of the Feast configuration file, with defaults
    preloaded

    """
    # Create the configuration file directory if needed
    config_dir = os.path.dirname(path)
    config_dir = config_dir.rstrip("/") + "/"

    os.makedirs(os.path.dirname(config_dir), exist_ok=True)

    # Create the configuration file itself
    config = ConfigParser(defaults=opt().defaults(), allow_no_value=True)
    if os.path.exists(path):
        config.read(path)

    # Store all configuration in a single section
    if not config.has_section(CONFIG_FILE_SECTION):
        config.add_section(CONFIG_FILE_SECTION)

    return config


def _get_feast_env_vars():
    """
    Get environmental variables that start with "FEAST_"

    Returns: Dict of Feast environmental variables (stripped of prefix)
    """
    feast_env_vars = {}
    for key in os.environ.keys():
        if key.upper().startswith(CONFIG_FEAST_ENV_VAR_PREFIX):
            feast_env_vars[key[len(CONFIG_FEAST_ENV_VAR_PREFIX) :]] = os.environ[key]
    return feast_env_vars


class Config:
    """
    Maintains and provides access to Feast configuration

    Configuration is stored as key/value pairs. The user can specify options
    through either input arguments to this class, environmental variables, or
    by setting the config in a configuration file

    """

    def __init__(
        self, options: Optional[Dict[str, str]] = None, path: Optional[str] = None,
    ):
        """
        Configuration options are returned as follows (higher replaces lower)
        1. Initialized options ("options" argument)
        2. Environmental variables (reloaded on every "get")
        3. Configuration file options (loaded once)
        4. Default options (loaded once from memory)

        Args:
            options: (optional) A list of initialized/hardcoded options.
            path: (optional) File path to configuration file
        """
        if not path:
            path = join(
                expanduser("~"),
                os.environ.get(FEAST_CONFIG_FILE_ENV, CONFIG_FILE_DEFAULT_DIRECTORY,),
                CONFIG_FILE_NAME,
            )

        config = _init_config(path)

        self._options = {}
        if options and isinstance(options, dict):
            self._options = options

        self._config = config  # type: ConfigParser
        self._path = path  # type: str

    def _get(self, option, default, get_method):
        fallback = {} if default is _UNSET else {"fallback": default}
        return get_method(
            CONFIG_FILE_SECTION,
            option,
            vars={**_get_feast_env_vars(), **self._options},
            **fallback,
        )

    def get(self, option, default=_UNSET):
        """
        Returns a single configuration option as a string

        Args:
            option: Name of the option
            default: Default value to return if option is not found

        Returns: String option that is returned

        """
        return self._get(option, default, self._config.get)

    def getboolean(self, option, default=_UNSET):
        """
         Returns a single configuration option as a boolean

         Args:
             option: Name of the option
             default: Default value to return if option is not found

         Returns: Boolean option value that is returned

         """
        return self._get(option, default, self._config.getboolean)

    def getint(self, option, default=_UNSET):
        """
         Returns a single configuration option as an integer

         Args:
             option: Name of the option
             default: Default value to return if option is not found

         Returns: Integer option value that is returned

         """
        return self._get(option, default, self._config.getint)

    def getfloat(self, option, default=_UNSET):
        """
         Returns a single configuration option as an integer

         Args:
             option: Name of the option
             default: Default value to return if option is not found

         Returns: Float option value that is returned

         """
        return self._get(option, default, self._config.getfloat)

    def set(self, option, value):
        """
        Sets a configuration option. Must be serializable to string
        Args:
            option: Option name to use as key
            value: Value to store under option
        """
        self._config.set(CONFIG_FILE_SECTION, option, value=str(value))

    def exists(self, option):
        """
        Tests whether a specific option is available

        Args:
            option: Name of the option to check

        Returns: Boolean true/false whether the option is set

        """
        try:
            self.get(option=option)
            return True
        except NoOptionError:
            return False

    def save(self):
        """
        Save the current configuration to disk. This does not include
        environmental variables or initialized options
        """
        defaults = self._config.defaults()
        try:
            self._config._defaults = {}
            self._config.write(open(self._path, "w"))
        finally:
            self._config._defaults = defaults

    def __str__(self):
        result = ""
        for section_name in self._config.sections():
            result += "\n[" + section_name + "]\n"
            for name, value in self._config.items(section_name):
                result += name + " = " + value + "\n"
        return result



import tempfile
from pathlib import Path
from textwrap import dedent

from feast.feature_store import FeatureStore
from tests.cli_utils import CliRunner
from tests.online_read_write_test import basic_rw_test


def test_workflow() -> None:
    """
    Test running apply on a sample repo, and make sure the infra gets created.
    """
    runner = CliRunner()
    with tempfile.TemporaryDirectory() as repo_dir_name, tempfile.TemporaryDirectory() as data_dir_name:

        # Construct an example repo in a temporary dir
        repo_path = Path(repo_dir_name)
        data_path = Path(data_dir_name)

        repo_config = repo_path / "feature_store.yaml"

        repo_config.write_text(
            dedent(
                f"""
        project: foo
        registry: {data_path / "registry.db"}
        provider: local
        online_store:
            path: {data_path / "online_store.db"}
        """
            )
        )

        repo_example = repo_path / "example.py"
        repo_example.write_text(
            (Path(__file__).parent / "example_feature_repo_1.py").read_text()
        )

        result = runner.run(["apply"], cwd=repo_path)
        assert result.returncode == 0

        # entity & feature view list commands should succeed
        result = runner.run(["entities", "list"], cwd=repo_path)
        assert result.returncode == 0
        result = runner.run(["feature-views", "list"], cwd=repo_path)
        assert result.returncode == 0

        # entity & feature view describe commands should succeed when objects exist
        result = runner.run(["entities", "describe", "driver"], cwd=repo_path)
        assert result.returncode == 0
        result = runner.run(
            ["feature-views", "describe", "driver_locations"], cwd=repo_path
        )
        assert result.returncode == 0

        # entity & feature view describe commands should fail when objects don't exist
        result = runner.run(["entities", "describe", "foo"], cwd=repo_path)
        assert result.returncode == 1
        result = runner.run(["feature-views", "describe", "foo"], cwd=repo_path)
        assert result.returncode == 1

        # Doing another apply should be a no op, and should not cause errors
        result = runner.run(["apply"], cwd=repo_path)
        assert result.returncode == 0

        basic_rw_test(
            FeatureStore(repo_path=str(repo_path), config=None),
            view_name="driver_locations",
        )

        result = runner.run(["teardown"], cwd=repo_path)
        assert result.returncode == 0


def test_non_local_feature_repo() -> None:
    """
    Test running apply on a sample repo, and make sure the infra gets created.
    """
    runner = CliRunner()
    with tempfile.TemporaryDirectory() as repo_dir_name:

        # Construct an example repo in a temporary dir
        repo_path = Path(repo_dir_name)

        repo_config = repo_path / "feature_store.yaml"

        repo_config.write_text(
            dedent(
                """
        project: foo
        registry: data/registry.db
        provider: local
        online_store:
            path: data/online_store.db
        """
            )
        )

        repo_example = repo_path / "example.py"
        repo_example.write_text(
            (Path(__file__).parent / "example_feature_repo_1.py").read_text()
        )

        result = runner.run(["apply"], cwd=repo_path)
        assert result.returncode == 0

        fs = FeatureStore(repo_path=str(repo_path))
        assert len(fs.list_feature_views()) == 3

        result = runner.run(["teardown"], cwd=repo_path)
        assert result.returncode == 0



import os
import sqlite3
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

import pandas as pd
import pytz

from feast import FeatureTable, utils
from feast.feature_view import FeatureView
from feast.infra.key_encoding_utils import serialize_entity_key
from feast.infra.offline_stores.helpers import get_offline_store_from_sources
from feast.infra.provider import (
    Provider,
    RetrievalJob,
    _convert_arrow_to_proto,
    _get_column_names,
    _run_field_mapping,
)
from feast.protos.feast.types.EntityKey_pb2 import EntityKey as EntityKeyProto
from feast.protos.feast.types.Value_pb2 import Value as ValueProto
from feast.registry import Registry
from feast.repo_config import RepoConfig, SqliteOnlineStoreConfig


class LocalProvider(Provider):
    _db_path: Path

    def __init__(self, config: RepoConfig, repo_path: Path):

        assert config is not None
        assert config.online_store is not None
        local_online_store_config = config.online_store
        assert isinstance(local_online_store_config, SqliteOnlineStoreConfig)
        local_path = Path(local_online_store_config.path)
        if local_path.is_absolute():
            self._db_path = local_path
        else:
            self._db_path = repo_path.joinpath(local_path)

    def _get_conn(self):
        Path(self._db_path).parent.mkdir(exist_ok=True)
        return sqlite3.connect(
            self._db_path, detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES
        )

    def update_infra(
        self,
        project: str,
        tables_to_delete: Sequence[Union[FeatureTable, FeatureView]],
        tables_to_keep: Sequence[Union[FeatureTable, FeatureView]],
        partial: bool,
    ):
        conn = self._get_conn()
        for table in tables_to_keep:
            conn.execute(
                f"CREATE TABLE IF NOT EXISTS {_table_id(project, table)} (entity_key BLOB, feature_name TEXT, value BLOB, event_ts timestamp, created_ts timestamp,  PRIMARY KEY(entity_key, feature_name))"
            )
            conn.execute(
                f"CREATE INDEX IF NOT EXISTS {_table_id(project, table)}_ek ON {_table_id(project, table)} (entity_key);"
            )

        for table in tables_to_delete:
            conn.execute(f"DROP TABLE IF EXISTS {_table_id(project, table)}")

    def teardown_infra(
        self, project: str, tables: Sequence[Union[FeatureTable, FeatureView]]
    ) -> None:
        os.unlink(self._db_path)

    def online_write_batch(
        self,
        project: str,
        table: Union[FeatureTable, FeatureView],
        data: List[
            Tuple[EntityKeyProto, Dict[str, ValueProto], datetime, Optional[datetime]]
        ],
        progress: Optional[Callable[[int], Any]],
    ) -> None:
        conn = self._get_conn()

        with conn:
            for entity_key, values, timestamp, created_ts in data:
                entity_key_bin = serialize_entity_key(entity_key)
                timestamp = _to_naive_utc(timestamp)
                if created_ts is not None:
                    created_ts = _to_naive_utc(created_ts)

                for feature_name, val in values.items():
                    conn.execute(
                        f"""
                            UPDATE {_table_id(project, table)}
                            SET value = ?, event_ts = ?, created_ts = ?
                            WHERE (entity_key = ? AND feature_name = ?)
                        """,
                        (
                            # SET
                            val.SerializeToString(),
                            timestamp,
                            created_ts,
                            # WHERE
                            entity_key_bin,
                            feature_name,
                        ),
                    )

                    conn.execute(
                        f"""INSERT OR IGNORE INTO {_table_id(project, table)}
                            (entity_key, feature_name, value, event_ts, created_ts)
                            VALUES (?, ?, ?, ?, ?)""",
                        (
                            entity_key_bin,
                            feature_name,
                            val.SerializeToString(),
                            timestamp,
                            created_ts,
                        ),
                    )

    def online_read(
        self,
        project: str,
        table: Union[FeatureTable, FeatureView],
        entity_keys: List[EntityKeyProto],
    ) -> List[Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]]:

        conn = self._get_conn()
        cur = conn.cursor()

        result: List[Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]] = []

        for entity_key in entity_keys:
            entity_key_bin = serialize_entity_key(entity_key)

            cur.execute(
                f"SELECT feature_name, value, event_ts FROM {_table_id(project, table)} WHERE entity_key = ?",
                (entity_key_bin,),
            )

            res = {}
            res_ts = None
            for feature_name, val_bin, ts in cur.fetchall():
                val = ValueProto()
                val.ParseFromString(val_bin)
                res[feature_name] = val
                res_ts = ts

            if not res:
                result.append((None, None))
            else:
                result.append((res_ts, res))
        return result

    def materialize_single_feature_view(
        self,
        feature_view: FeatureView,
        start_date: datetime,
        end_date: datetime,
        registry: Registry,
        project: str,
    ) -> None:
        entities = []
        for entity_name in feature_view.entities:
            entities.append(registry.get_entity(entity_name, project))

        (
            join_key_columns,
            feature_name_columns,
            event_timestamp_column,
            created_timestamp_column,
        ) = _get_column_names(feature_view, entities)

        start_date = utils.make_tzaware(start_date)
        end_date = utils.make_tzaware(end_date)

        offline_store = get_offline_store_from_sources([feature_view.input])
        table = offline_store.pull_latest_from_table_or_query(
            data_source=feature_view.input,
            join_key_columns=join_key_columns,
            feature_name_columns=feature_name_columns,
            event_timestamp_column=event_timestamp_column,
            created_timestamp_column=created_timestamp_column,
            start_date=start_date,
            end_date=end_date,
        )

        if feature_view.input.field_mapping is not None:
            table = _run_field_mapping(table, feature_view.input.field_mapping)

        join_keys = [entity.join_key for entity in entities]
        rows_to_write = _convert_arrow_to_proto(table, feature_view, join_keys)

        self.online_write_batch(project, feature_view, rows_to_write, None)

        feature_view.materialization_intervals.append((start_date, end_date))
        registry.apply_feature_view(feature_view, project)

    @staticmethod
    def get_historical_features(
        config: RepoConfig,
        feature_views: List[FeatureView],
        feature_refs: List[str],
        entity_df: Union[pd.DataFrame, str],
        registry: Registry,
        project: str,
    ) -> RetrievalJob:
        offline_store = get_offline_store_from_sources(
            [feature_view.input for feature_view in feature_views]
        )
        return offline_store.get_historical_features(
            config=config,
            feature_views=feature_views,
            feature_refs=feature_refs,
            entity_df=entity_df,
            registry=registry,
            project=project,
        )


def _table_id(project: str, table: Union[FeatureTable, FeatureView]) -> str:
    return f"{project}_{table.name}"


def _to_naive_utc(ts: datetime):
    if ts.tzinfo is None:
        return ts
    else:
        return ts.astimezone(pytz.utc).replace(tzinfo=None)



# Copyright 2019 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from datetime import datetime
from pathlib import Path
from typing import List

import click
import pkg_resources
import yaml

from feast.errors import FeastObjectNotFoundException, FeastProviderLoginError
from feast.feature_store import FeatureStore
from feast.repo_config import load_repo_config
from feast.repo_operations import (
    apply_total,
    cli_check_repo,
    generate_project_name,
    init_repo,
    registry_dump,
    teardown,
)

_logger = logging.getLogger(__name__)
DATETIME_ISO = "%Y-%m-%dT%H:%M:%s"


@click.group()
def cli():
    """
    Feast CLI

    For more information, see our public docs at https://docs.feast.dev/

    For any questions, you can reach us at https://slack.feast.dev/
    """
    pass


@cli.command()
def version():
    """
    Display Feast SDK version
    """
    print(f'Feast SDK Version: "{pkg_resources.get_distribution("feast")}"')


@cli.group(name="entities")
def entities_cmd():
    """
    Access entities
    """
    pass


@entities_cmd.command("describe")
@click.argument("name", type=click.STRING)
def entity_describe(name: str):
    """
    Describe an entity
    """
    cli_check_repo(Path.cwd())
    store = FeatureStore(repo_path=str(Path.cwd()))

    try:
        entity = store.get_entity(name)
    except FeastObjectNotFoundException as e:
        print(e)
        exit(1)

    print(
        yaml.dump(
            yaml.safe_load(str(entity)), default_flow_style=False, sort_keys=False
        )
    )


@entities_cmd.command(name="list")
def entity_list():
    """
    List all entities
    """
    cli_check_repo(Path.cwd())
    store = FeatureStore(repo_path=str(Path.cwd()))
    table = []
    for entity in store.list_entities():
        table.append([entity.name, entity.description, entity.value_type])

    from tabulate import tabulate

    print(tabulate(table, headers=["NAME", "DESCRIPTION", "TYPE"], tablefmt="plain"))


@cli.group(name="feature-views")
def feature_views_cmd():
    """
    Access feature views
    """
    pass


@feature_views_cmd.command("describe")
@click.argument("name", type=click.STRING)
def feature_view_describe(name: str):
    """
    Describe a feature view
    """
    cli_check_repo(Path.cwd())
    store = FeatureStore(repo_path=str(Path.cwd()))

    try:
        feature_view = store.get_feature_view(name)
    except FeastObjectNotFoundException as e:
        print(e)
        exit(1)

    print(
        yaml.dump(
            yaml.safe_load(str(feature_view)), default_flow_style=False, sort_keys=False
        )
    )


@feature_views_cmd.command(name="list")
def feature_view_list():
    """
    List all feature views
    """
    cli_check_repo(Path.cwd())
    store = FeatureStore(repo_path=str(Path.cwd()))
    table = []
    for feature_view in store.list_feature_views():
        table.append([feature_view.name, feature_view.entities])

    from tabulate import tabulate

    print(tabulate(table, headers=["NAME", "ENTITIES"], tablefmt="plain"))


@cli.command("apply")
def apply_total_command():
    """
    Create or update a feature store deployment
    """
    cli_check_repo(Path.cwd())
    repo_config = load_repo_config(Path.cwd())
    try:
        apply_total(repo_config, Path.cwd())
    except FeastProviderLoginError as e:
        print(str(e))


@cli.command("teardown")
def teardown_command():
    """
    Tear down deployed feature store infrastructure
    """
    cli_check_repo(Path.cwd())
    repo_config = load_repo_config(Path.cwd())

    teardown(repo_config, Path.cwd())


@cli.command("registry-dump")
def registry_dump_command():
    """
    Print contents of the metadata registry
    """
    cli_check_repo(Path.cwd())
    repo_config = load_repo_config(Path.cwd())

    registry_dump(repo_config, repo_path=Path.cwd())


@cli.command("materialize")
@click.argument("start_ts")
@click.argument("end_ts")
@click.option(
    "--views", "-v", help="Feature views to materialize", multiple=True,
)
def materialize_command(start_ts: str, end_ts: str, views: List[str]):
    """
    Run a (non-incremental) materialization job to ingest data into the online store. Feast
    will read all data between START_TS and END_TS from the offline store and write it to the
    online store. If you don't specify feature view names using --views, all registered Feature
    Views will be materialized.

    START_TS and END_TS should be in ISO 8601 format, e.g. '2021-07-16T19:20:01'
    """
    cli_check_repo(Path.cwd())
    store = FeatureStore(repo_path=str(Path.cwd()))
    store.materialize(
        feature_views=None if not views else views,
        start_date=datetime.fromisoformat(start_ts),
        end_date=datetime.fromisoformat(end_ts),
    )


@cli.command("materialize-incremental")
@click.argument("end_ts")
@click.option(
    "--views", "-v", help="Feature views to incrementally materialize", multiple=True,
)
def materialize_incremental_command(end_ts: str, views: List[str]):
    """
    Run an incremental materialization job to ingest new data into the online store. Feast will read
    all data from the previously ingested point to END_TS from the offline store and write it to the
    online store. If you don't specify feature view names using --views, all registered Feature
    Views will be incrementally materialized.

    END_TS should be in ISO 8601 format, e.g. '2021-07-16T19:20:01'
    """
    cli_check_repo(Path.cwd())
    store = FeatureStore(repo_path=str(Path.cwd()))
    store.materialize_incremental(
        feature_views=None if not views else views,
        end_date=datetime.fromisoformat(end_ts),
    )


@cli.command("init")
@click.argument("PROJECT_DIRECTORY", required=False)
@click.option(
    "--minimal", "-m", is_flag=True, help="Create an empty project repository"
)
@click.option(
    "--template",
    "-t",
    type=click.Choice(["local", "gcp"], case_sensitive=False),
    help="Specify a template for the created project",
    default="local",
)
def init_command(project_directory, minimal: bool, template: str):
    """Create a new Feast repository"""
    if not project_directory:
        project_directory = generate_project_name()

    if minimal:
        template = "minimal"

    init_repo(project_directory, template)


if __name__ == "__main__":
    cli()



import time
from dataclasses import asdict, dataclass
from datetime import datetime, timedelta
from typing import List, Optional, Union

import pandas
import pyarrow
from google.auth.exceptions import DefaultCredentialsError
from google.cloud import bigquery
from jinja2 import BaseLoader, Environment

from feast.data_source import BigQuerySource, DataSource
from feast.errors import FeastProviderLoginError
from feast.feature_view import FeatureView
from feast.infra.offline_stores.offline_store import OfflineStore
from feast.infra.provider import (
    RetrievalJob,
    _get_requested_feature_views_to_features_dict,
)
from feast.registry import Registry
from feast.repo_config import RepoConfig


class BigQueryOfflineStore(OfflineStore):
    @staticmethod
    def pull_latest_from_table_or_query(
        data_source: DataSource,
        join_key_columns: List[str],
        feature_name_columns: List[str],
        event_timestamp_column: str,
        created_timestamp_column: Optional[str],
        start_date: datetime,
        end_date: datetime,
    ) -> pyarrow.Table:
        assert isinstance(data_source, BigQuerySource)
        from_expression = data_source.get_table_query_string()

        partition_by_join_key_string = ", ".join(join_key_columns)
        if partition_by_join_key_string != "":
            partition_by_join_key_string = (
                "PARTITION BY " + partition_by_join_key_string
            )
        timestamps = [event_timestamp_column]
        if created_timestamp_column:
            timestamps.append(created_timestamp_column)
        timestamp_desc_string = " DESC, ".join(timestamps) + " DESC"
        field_string = ", ".join(join_key_columns + feature_name_columns + timestamps)

        query = f"""
            SELECT {field_string}
            FROM (
                SELECT {field_string},
                ROW_NUMBER() OVER({partition_by_join_key_string} ORDER BY {timestamp_desc_string}) AS _feast_row
                FROM {from_expression}
                WHERE {event_timestamp_column} BETWEEN TIMESTAMP('{start_date}') AND TIMESTAMP('{end_date}')
            )
            WHERE _feast_row = 1
            """

        return BigQueryOfflineStore._pull_query(query)

    @staticmethod
    def _pull_query(query: str) -> pyarrow.Table:
        client = _get_bigquery_client()
        query_job = client.query(query)
        return query_job.to_arrow()

    @staticmethod
    def get_historical_features(
        config: RepoConfig,
        feature_views: List[FeatureView],
        feature_refs: List[str],
        entity_df: Union[pandas.DataFrame, str],
        registry: Registry,
        project: str,
    ) -> RetrievalJob:
        # TODO: Add entity_df validation in order to fail before interacting with BigQuery

        client = _get_bigquery_client()

        if type(entity_df) is str:
            entity_df_sql_table = f"({entity_df})"
        elif isinstance(entity_df, pandas.DataFrame):
            if "event_timestamp" not in entity_df.columns:
                raise ValueError(
                    "Please provide an entity_df with a column named event_timestamp representing the time of events."
                )
            table_id = _upload_entity_df_into_bigquery(
                config.project, entity_df, client
            )
            entity_df_sql_table = f"`{table_id}`"
        else:
            raise ValueError(
                f"The entity dataframe you have provided must be a Pandas DataFrame or BigQuery SQL query, "
                f"but we found: {type(entity_df)} "
            )

        # Build a query context containing all information required to template the BigQuery SQL query
        query_context = get_feature_view_query_context(
            feature_refs, feature_views, registry, project
        )

        # TODO: Infer min_timestamp and max_timestamp from entity_df
        # Generate the BigQuery SQL query from the query context
        query = build_point_in_time_query(
            query_context,
            min_timestamp=datetime.now() - timedelta(days=365),
            max_timestamp=datetime.now() + timedelta(days=1),
            left_table_query_string=entity_df_sql_table,
        )

        job = BigQueryRetrievalJob(query=query, client=client)
        return job


class BigQueryRetrievalJob(RetrievalJob):
    def __init__(self, query, client):
        self.query = query
        self.client = client

    def to_df(self):
        # TODO: Ideally only start this job when the user runs "get_historical_features", not when they run to_df()
        df = self.client.query(self.query).to_dataframe(create_bqstorage_client=True)
        return df


@dataclass(frozen=True)
class FeatureViewQueryContext:
    """Context object used to template a BigQuery point-in-time SQL query"""

    name: str
    ttl: int
    entities: List[str]
    features: List[str]  # feature reference format
    table_ref: str
    event_timestamp_column: str
    created_timestamp_column: Optional[str]
    query: str
    table_subquery: str
    entity_selections: List[str]


def _upload_entity_df_into_bigquery(project, entity_df, client) -> str:
    """Uploads a Pandas entity dataframe into a BigQuery table and returns a reference to the resulting table"""

    # First create the BigQuery dataset if it doesn't exist
    dataset = bigquery.Dataset(f"{client.project}.feast_{project}")
    dataset.location = "US"
    client.create_dataset(
        dataset, exists_ok=True
    )  # TODO: Consider moving this to apply or BigQueryOfflineStore

    # Drop the index so that we dont have unnecessary columns
    entity_df.reset_index(drop=True, inplace=True)

    # Upload the dataframe into BigQuery, creating a temporary table
    job_config = bigquery.LoadJobConfig()
    table_id = f"{client.project}.feast_{project}.entity_df_{int(time.time())}"
    job = client.load_table_from_dataframe(entity_df, table_id, job_config=job_config,)
    job.result()

    # Ensure that the table expires after some time
    table = client.get_table(table=table_id)
    table.expires = datetime.utcnow() + timedelta(minutes=30)
    client.update_table(table, ["expires"])

    return table_id


def get_feature_view_query_context(
    feature_refs: List[str],
    feature_views: List[FeatureView],
    registry: Registry,
    project: str,
) -> List[FeatureViewQueryContext]:
    """Build a query context containing all information required to template a BigQuery point-in-time SQL query"""

    feature_views_to_feature_map = _get_requested_feature_views_to_features_dict(
        feature_refs, feature_views
    )

    query_context = []
    for feature_view, features in feature_views_to_feature_map.items():
        join_keys = []
        entity_selections = []
        reverse_field_mapping = {
            v: k for k, v in feature_view.input.field_mapping.items()
        }
        for entity_name in feature_view.entities:
            entity = registry.get_entity(entity_name, project)
            join_keys.append(entity.join_key)
            join_key_column = reverse_field_mapping.get(
                entity.join_key, entity.join_key
            )
            entity_selections.append(f"{join_key_column} AS {entity.join_key}")

        if isinstance(feature_view.ttl, timedelta):
            ttl_seconds = int(feature_view.ttl.total_seconds())
        else:
            ttl_seconds = 0

        assert isinstance(feature_view.input, BigQuerySource)

        event_timestamp_column = feature_view.input.event_timestamp_column
        created_timestamp_column = feature_view.input.created_timestamp_column

        context = FeatureViewQueryContext(
            name=feature_view.name,
            ttl=ttl_seconds,
            entities=join_keys,
            features=features,
            table_ref=feature_view.input.table_ref,
            event_timestamp_column=reverse_field_mapping.get(
                event_timestamp_column, event_timestamp_column
            ),
            created_timestamp_column=reverse_field_mapping.get(
                created_timestamp_column, created_timestamp_column
            ),
            # TODO: Make created column optional and not hardcoded
            query=feature_view.input.query,
            table_subquery=feature_view.input.get_table_query_string(),
            entity_selections=entity_selections,
        )
        query_context.append(context)
    return query_context


def build_point_in_time_query(
    feature_view_query_contexts: List[FeatureViewQueryContext],
    min_timestamp: datetime,
    max_timestamp: datetime,
    left_table_query_string: str,
):
    """Build point-in-time query between each feature view table and the entity dataframe"""
    template = Environment(loader=BaseLoader()).from_string(
        source=SINGLE_FEATURE_VIEW_POINT_IN_TIME_JOIN
    )

    # Add additional fields to dict
    template_context = {
        "min_timestamp": min_timestamp,
        "max_timestamp": max_timestamp,
        "left_table_query_string": left_table_query_string,
        "featureviews": [asdict(context) for context in feature_view_query_contexts],
    }

    query = template.render(template_context)
    return query


def _get_bigquery_client():
    try:
        from google.cloud import bigquery

        client = bigquery.Client()
    except DefaultCredentialsError as e:
        raise FeastProviderLoginError(
            str(e)
            + '\nIt may be necessary to run "gcloud auth application-default login" if you would like to use your '
            "local Google Cloud account"
        )
    except EnvironmentError as e:
        raise FeastProviderLoginError(
            "GCP error: "
            + str(e)
            + "\nIt may be necessary to set a default GCP project by running "
            '"gcloud config set project your-project"'
        )

    return client


# TODO: Optimizations
#   * Use GENERATE_UUID() instead of ROW_NUMBER(), or join on entity columns directly
#   * Precompute ROW_NUMBER() so that it doesn't have to be recomputed for every query on entity_dataframe
#   * Create temporary tables instead of keeping all tables in memory

SINGLE_FEATURE_VIEW_POINT_IN_TIME_JOIN = """
WITH entity_dataframe AS (
    SELECT ROW_NUMBER() OVER() AS row_number, edf.* FROM {{ left_table_query_string }} as edf
),
{% for featureview in featureviews %}
/*
 This query template performs the point-in-time correctness join for a single feature set table
 to the provided entity table.
 1. Concatenate the timestamp and entities from the feature set table with the entity dataset.
 Feature values are joined to this table later for improved efficiency.
 featureview_timestamp is equal to null in rows from the entity dataset.
 */
{{ featureview.name }}__union_features AS (
SELECT
  -- unique identifier for each row in the entity dataset.
  row_number,
  -- event_timestamp contains the timestamps to join onto
  event_timestamp,
  -- the feature_timestamp, i.e. the latest occurrence of the requested feature relative to the entity_dataset timestamp
  NULL as {{ featureview.name }}_feature_timestamp,
  -- created timestamp of the feature at the corresponding feature_timestamp
  {{ 'NULL as created_timestamp,' if featureview.created_timestamp_column else '' }}
  -- select only entities belonging to this feature set
  {{ featureview.entities | join(', ')}},
  -- boolean for filtering the dataset later
  true AS is_entity_table
FROM entity_dataframe
UNION ALL
SELECT
  NULL as row_number,
  {{ featureview.event_timestamp_column }} as event_timestamp,
  {{ featureview.event_timestamp_column }} as {{ featureview.name }}_feature_timestamp,
  {{ featureview.created_timestamp_column ~ ' as created_timestamp,' if featureview.created_timestamp_column else '' }}
  {{ featureview.entity_selections | join(', ')}},
  false AS is_entity_table
FROM {{ featureview.table_subquery }} WHERE {{ featureview.event_timestamp_column }} <= '{{ max_timestamp }}'
{% if featureview.ttl == 0 %}{% else %}AND {{ featureview.event_timestamp_column }} >= Timestamp_sub(TIMESTAMP '{{ min_timestamp }}', interval {{ featureview.ttl }} second){% endif %}
),
/*
 2. Window the data in the unioned dataset, partitioning by entity and ordering by event_timestamp, as
 well as is_entity_table.
 Within each window, back-fill the feature_timestamp - as a result of this, the null feature_timestamps
 in the rows from the entity table should now contain the latest timestamps relative to the row's
 event_timestamp.
 For rows where event_timestamp(provided datetime) - feature_timestamp > max age, set the
 feature_timestamp to null.
 */
{{ featureview.name }}__joined AS (
SELECT
  row_number,
  event_timestamp,
  {{ featureview.entities | join(', ')}},
  {% for feature in featureview.features %}
  IF(event_timestamp >= {{ featureview.name }}_feature_timestamp {% if featureview.ttl == 0 %}{% else %}AND Timestamp_sub(event_timestamp, interval {{ featureview.ttl }} second) < {{ featureview.name }}_feature_timestamp{% endif %}, {{ featureview.name }}__{{ feature }}, NULL) as {{ featureview.name }}__{{ feature }}{% if loop.last %}{% else %}, {% endif %}
  {% endfor %}
FROM (
SELECT
  row_number,
  event_timestamp,
  {{ featureview.entities | join(', ')}},
  {{ 'FIRST_VALUE(created_timestamp IGNORE NULLS) over w AS created_timestamp,' if featureview.created_timestamp_column else '' }}
  FIRST_VALUE({{ featureview.name }}_feature_timestamp IGNORE NULLS) over w AS {{ featureview.name }}_feature_timestamp,
  is_entity_table
FROM {{ featureview.name }}__union_features
WINDOW w AS (PARTITION BY {{ featureview.entities | join(', ') }} ORDER BY event_timestamp DESC, is_entity_table DESC{{', created_timestamp DESC' if featureview.created_timestamp_column else ''}} ROWS BETWEEN CURRENT ROW AND UNBOUNDED FOLLOWING)
)
/*
 3. Select only the rows from the entity table, and join the features from the original feature set table
 to the dataset using the entity values, feature_timestamp, and created_timestamps.
 */
LEFT JOIN (
SELECT
  {{ featureview.event_timestamp_column }} as {{ featureview.name }}_feature_timestamp,
  {{ featureview.created_timestamp_column ~ ' as created_timestamp,' if featureview.created_timestamp_column else '' }}
  {{ featureview.entity_selections | join(', ')}},
  {% for feature in featureview.features %}
  {{ feature }} as {{ featureview.name }}__{{ feature }}{% if loop.last %}{% else %}, {% endif %}
  {% endfor %}
FROM {{ featureview.table_subquery }} WHERE {{ featureview.event_timestamp_column }} <= '{{ max_timestamp }}'
{% if featureview.ttl == 0 %}{% else %}AND {{ featureview.event_timestamp_column }} >= Timestamp_sub(TIMESTAMP '{{ min_timestamp }}', interval {{ featureview.ttl }} second){% endif %}
) USING ({{ featureview.name }}_feature_timestamp,{{ ' created_timestamp,' if featureview.created_timestamp_column else '' }} {{ featureview.entities | join(', ')}})
WHERE is_entity_table
),
/*
 4. Finally, deduplicate the rows by selecting the first occurrence of each entity table row_number.
 */
{{ featureview.name }}__deduped AS (SELECT
  k.*
FROM (
  SELECT ARRAY_AGG(row LIMIT 1)[OFFSET(0)] k
  FROM {{ featureview.name }}__joined row
  GROUP BY row_number
)){% if loop.last %}{% else %}, {% endif %}

{% endfor %}
/*
 Joins the outputs of multiple time travel joins to a single table.
 */
SELECT edf.event_timestamp as event_timestamp, * EXCEPT (row_number, event_timestamp) FROM entity_dataframe edf
{% for featureview in featureviews %}
LEFT JOIN (
    SELECT
    row_number,
    {% for feature in featureview.features %}
    {{ featureview.name }}__{{ feature }}{% if loop.last %}{% else %}, {% endif %}
    {% endfor %}
    FROM {{ featureview.name }}__deduped
) USING (row_number)
{% endfor %}
ORDER BY event_timestamp
"""
